@@ -1,6 +1,7 @@
 from collections import defaultdict
 import json
 import logging
+import statistics
 from typing import Dict, List, Tuple
 
 import numpy as np
@@ -24,6 +25,9 @@ MISSING_CATEGORICAL_CATEGORY = "null"
 DECISION_FEATURE_NAME = "decision"
 POSITION_FEATURE_NAME = "position"
 
+NUMPY_NUMERIC_TYPE = np.number
+NUMPY_CATEGORICAL_TYPE = np.str_
+
 
 def get_preprocess_feature_order(
     features_spec: Dict, features_to_use: List[str]
@@ -74,6 +78,15 @@ def preprocess_feature(
     values = values.reshape(-1, 1)
 
     if feature_type == "N":
+        # numeric features should be... numbers
+        if not np.issubdtype(values.dtype, NUMPY_NUMERIC_TYPE):
+            logger.warning(
+                f"{feature_name} of type `N` has values of type {values.dtype}, "
+                "but it should have numeric values. Attempting to cast it, but "
+                "this might cause issues..."
+            )
+            values = values.astype(NUMPY_NUMERIC_TYPE)
+
         # standard scaler seems to be right choice for numeric features
         # in regression problems
         # http://rajeshmahajan.com/standard-scaler-v-min-max-scaler-machine-learning/
@@ -85,9 +98,17 @@ def preprocess_feature(
         values = preprocessor.fit_transform(values)
         df = pd.DataFrame(values.squeeze(), columns=[feature_name])
     elif feature_type == "C":
+
+        # categorical features should be strings
+        if not values.dtype.type is NUMPY_CATEGORICAL_TYPE:
+            logger.warning(
+                f"{feature_name} of type `C` has values of type {values.dtype}, "
+                "but it should have string values. Attempting to cast it, but "
+                "this might cause issues..."
+            )
+            values = values.astype(NUMPY_CATEGORICAL_TYPE)
+
         imputer = None
-        possible_values = set(values.squeeze().tolist())
-        logger.info(f"{prefix}{feature_name} [{feature_type}]: {possible_values}")
         preprocessor = preprocessing.OneHotEncoder(sparse=False)
         # always add "null" as a possible value to categorical features
         # so a missing value is fine during inference even if a missing
@@ -97,6 +118,9 @@ def preprocess_feature(
         else:
             fit_values = np.append(values, [[MISSING_CATEGORICAL_CATEGORY]], axis=0)
 
+        possible_values = set(fit_values.squeeze().tolist())
+        logger.info(f"{prefix}{feature_name} [{feature_type}]: {possible_values}")
+
         preprocessor.fit(fit_values)
         values = preprocessor.transform(values)
         preprocessor.col_names = [
@@ -111,8 +135,8 @@ def preprocess_feature(
 
 def preprocess_data(
     raw_data: pd.DataFrame,
-    reward_function: Dict,
     experiment_params: Dict,
+    reward_type: str,
     features_to_use: List[str] = ["*"],
     dense_features_to_use: List[str] = ["*"],
     shuffle_data: bool = True,
@@ -128,39 +152,8 @@ def preprocess_data(
     # load the json string into json objects and expand into columns
     X = pd.json_normalize(raw_data["context"].apply(json.loads))
     X[DECISION_FEATURE_NAME] = raw_data[DECISION_FEATURE_NAME].values
+    X["reward"] = raw_data["reward"].astype(float).values
 
-    # fill in missing (NaN) rewards with empty metric maps
-    raw_data["immediate_reward"] = raw_data["immediate_reward"].fillna("{}")
-    immediate_y = pd.json_normalize(
-        raw_data["immediate_reward"].apply(lambda x: json.loads(x))
-    )
-    immediate_y = immediate_y.fillna(0)
-
-    # calculate the delayed rewards which requires joining them to their
-    # previous decisions
-    raw_data["end_of_mdp_reward"] = raw_data["delayed_reward"].fillna("{}")
-    X["end_of_mdp_reward"] = (
-        raw_data["end_of_mdp_reward"].apply(lambda x: json.loads(x)).values
-    )
-    delayed_rewards = X.apply(
-        lambda row: row["end_of_mdp_reward"].get(str(row[DECISION_FEATURE_NAME]), {}),
-        axis=1,
-    )
-    delayed_y = pd.json_normalize(delayed_rewards)
-    delayed_y = delayed_y.fillna(0)
-
-    # initialize the final reward scalar with 0s
-    X["reward"] = np.zeros(len(X))
-
-    # then add the linear combination of immediate rewards
-    for metrics_name, series in immediate_y.iteritems():
-        X["reward"] += series * reward_function[metrics_name]
-
-    # then add the linear combination of delayed rewards
-    for metrics_name, series in delayed_y.iteritems():
-        X["reward"] += series * reward_function[metrics_name]
-
-    reward_df = X["reward"]
     float_feature_df = pd.DataFrame()
     id_list_feature_df = pd.DataFrame()
 
@@ -174,7 +167,14 @@ def preprocess_data(
         f"{non_zero_reward_rows} of {total_rows} rows have non-zero reward "
         f"({percent_non_zero}%)"
     )
-    logger.info(f"Reward range: [{min(X['reward'])}, {max(X['reward'])}]")
+
+    max_reward = max(X["reward"])
+    min_reward = min(X["reward"])
+    logger.info(f"Reward range: [{min_reward}, {max_reward}]")
+    if reward_type == "binary" and max_reward > 1:
+        logger.warning("Reward type specified as binary, but reward > 1 found.")
+        logger.warning("Coverting rewards to binary values.")
+        X["reward"] = (X["reward"] > 0) * 1
 
     logger.info("Manually selected features")
     logger.info(_sub_dividing_text())
@@ -221,23 +221,42 @@ def preprocess_data(
             dense = defaultdict(list)
             # TODO: don't like that this is O(n^2), think about better way to do this
             for val in X[feature_name].values:
-                dense_features = product_set_meta["dense"].get(val)
-                if not dense_features:
-                    logger.warning(
-                        f"No dense representation found for '{feature_name}'"
-                        f" product set value '{val}'."
-                    )
-                for idx, feature_spec in enumerate(product_set_meta["features"]):
+                if not isinstance(val, list):
+                    val = [val]
 
-                    dense_feature_name = feature_spec["name"]
+                dense_matrix = []
+                for v in val:
+                    dense_features = product_set_meta["dense"].get(v)
                     if not dense_features:
-                        dense_feature_val = (
-                            MISSING_CATEGORICAL_CATEGORY
-                            if feature_spec["type"] == "C"
-                            else None
+                        logger.warning(
+                            f"No dense representation found for '{feature_name}'"
+                            f" product set value '{v}'."
                         )
                     else:
-                        dense_feature_val = dense_features[idx]
+                        dense_matrix.append(dense_features)
+
+                if not dense_matrix:
+                    # there were no or no valid id features to add, add an
+                    # empty row to be imputed
+                    dense_matrix.append([])
+
+                for idx, feature_spec in enumerate(product_set_meta["features"]):
+                    dense_feature_name = feature_spec["name"]
+                    row_vals = []
+                    for row in dense_matrix:
+                        if not row:
+                            dense_feature_val = (
+                                MISSING_CATEGORICAL_CATEGORY
+                                if feature_spec["type"] == "C"
+                                else None
+                            )
+                        else:
+                            dense_feature_val = row[idx]
+                        row_vals.append(dense_feature_val)
+
+                    dense_feature_val = flatten_dense_id_list_feature(
+                        row_vals, feature_spec["type"]
+                    )
                     dense[dense_feature_name].append(dense_feature_val)
 
             for idx, feature_spec in enumerate(product_set_meta["features"]):
@@ -252,7 +271,9 @@ def preprocess_data(
                     continue
 
                 dtype = (
-                    np.dtype(float) if feature_spec["type"] == "N" else np.dtype(object)
+                    NUMPY_NUMERIC_TYPE
+                    if feature_spec["type"] == "N"
+                    else NUMPY_CATEGORICAL_TYPE
                 )
 
                 vals = dense[feature_spec["name"]]
@@ -279,16 +300,30 @@ def preprocess_data(
             product_set_id = experiment_params["features"][feature_name][
                 "product_set_id"
             ]
+
             str_to_int_map = id_feature_str_to_int_map[product_set_id]
+
+            def _str_id_list_converter(input):
+                if not isinstance(input, list):
+                    input = [input]
+                out = []
+                for i in input:
+                    int_id = str_to_int_map.get(i)
+                    if not int_id:
+                        logger.info(f"{i} missing from product_set {product_set_id}")
+                    else:
+                        out.append(int_id)
+                return out
+
             final_id_feature_order.append(feature_name)
             id_list_feature_df[feature_name] = pd.Series(X[feature_name].values).apply(
-                lambda x: str_to_int_map[x]
+                lambda x: _str_id_list_converter(x)
             )
             transforms[feature_name] = None
             imputers[feature_name] = None
 
     return {
-        "y": reward_df,
+        "y": X["reward"],
         "X_float": float_feature_df,
         "X_id_list": id_list_feature_df,
         "transforms": transforms,
@@ -301,6 +336,22 @@ def preprocess_data(
     }
 
 
+def flatten_dense_id_list_feature(vals, feature_type):
+    if feature_type == "C":
+        # flatten categorical dense features by taking the mode
+        return max(set(vals), key=vals.count)
+    elif feature_type == "N":
+        # drop nones and take mean across values remaining
+        vals = [v for v in vals if v is not None]
+
+        if not vals:
+            return None
+
+        return statistics.mean(vals)
+
+    raise Exception(f"Feature type {feature_type} not supported.")
+
+
 def data_to_pytorch(data: Dict):
     X = {}
 
